﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CheckHammingCodesAndPageParities
{
    //Version 1.1
    class Program
    {
        //Information about the format in which the FDR records the data can be found in the post 
        //at http://forums.randi.org/showpost.php?p=2022420&postcount=85 and in figure 4.2-2 of 
        //the document http://www.asc.gov.tw/author_files/ASC-TRT-99-007.pdf

        //The words in the figure are 16 bit words stored Least Significant Byte first and the 
        //Hamming Code & Page Parity occupies the last 11 bits of each 128 byte page. Within each 
        //16 bit word, the most significant bit is the first bit in the bit stream. 
        static void Main(string[] args)
        {
            StreamWriter errorsOutputFile = new StreamWriter(@"C:\Users\Admin\Documents\Pilots For 911 Truth\Hamming Codes and Page Parities.txt");
            StreamWriter headersOutputFile = new StreamWriter(@"C:\Users\Admin\Documents\Pilots For 911 Truth\FDR File Headers.txt");
            DirectoryInfo FDRDirectoryInfo = new DirectoryInfo(@"C:\Users\Admin\Documents\Pilots For 911 Truth\FDR Files\");
            foreach (FileInfo FDRFile in FDRDirectoryInfo.GetFiles())
            {
                FileStream inputFile = new FileStream(FDRFile.FullName, FileMode.Open, FileAccess.Read);
                int fileLength = (int)inputFile.Length;
                int fileTextHeaderLength = 0;
                //Search through the text header at the beginning of the file until the first header 
                //or erased page is found.
                string headerText = "";
                for (; fileTextHeaderLength < fileLength; fileTextHeaderLength += 2)
                {
                    byte[] b = new byte[2];
                    inputFile.Read(b, 0, 2);
                    int w = (b[1] << 8) | b[0];
                    if ((w == 0xFE6B) || (w == 0xFFFF))
                        break;
                    for (int i = 0; i < b.Length; i++)
                    {
                        char c = (char)b[i];
                        headerText += c;
                    }
                }
                headersOutputFile.WriteLine("Text header for file {0}:", FDRFile.Name);
                headersOutputFile.WriteLine(headerText);
                inputFile.Seek(fileTextHeaderLength, SeekOrigin.Begin);
                List<string> incorrectHammingCodeAndPageParityPages = new List<string>();
                int pageNum = 0;
                for (int filePos = fileTextHeaderLength; filePos < fileLength; filePos += 128)
                {
                    byte[] page = new byte[128];
                    inputFile.Read(page, 0, 128);
                    //Only check flight data not header pages
                    if ((pageNum++ & 0x1FF) >= 2)
                    {
                        int pageParity = 0;
                        int hammingCode = 0;
                        int shiftNum = 0;
                        //Look through each bit of the page except the page parity, one at a time.
                        for (int bitNum0 = 0; bitNum0 < 1023; bitNum0++)
                        {
                            int bitNum1 = bitNum0 + 1;
                            int bitHammingCode = (bitNum1 == (1 << shiftNum)) ? 1023 - shiftNum++ : bitNum1;
                            if (((page[(bitNum0 >> 3) ^ 1] << (bitNum0 & 7)) & 0x80) != 0)
                            {
                                //The bit numbered BitNum0 of the page is 1, so update the page Hamming code and page parity.
                                if (bitNum0 < 1013)
                                    hammingCode ^= bitHammingCode;
                                pageParity ^= 1;
                            }
                        }
                        string calculatedHammingCodeAndPageParity = ((hammingCode << 1) | pageParity).ToString("X");
                        string fileHammingCodeAndPageParity = (((page[127] << 8) | page[126]) & 0x7FF).ToString("X");
                        if (calculatedHammingCodeAndPageParity != fileHammingCodeAndPageParity)
                        {
                            string line = string.Format("    Calculated Hamming code and page parity of {0}(hex) but found {1}(hex) in page starting at file offset {2:X}(hex)", calculatedHammingCodeAndPageParity, fileHammingCodeAndPageParity, filePos);
                            incorrectHammingCodeAndPageParityPages.Add(line);
                        }
                    }
                }
                inputFile.Close();
                int fileDataLength = fileLength - fileTextHeaderLength;
                errorsOutputFile.WriteLine("File {0}: {1} bytes of text header, {2} bytes of data, {3} errors found", FDRFile.Name, fileTextHeaderLength, fileDataLength, incorrectHammingCodeAndPageParityPages.Count);
                foreach (string line in incorrectHammingCodeAndPageParityPages)
                    errorsOutputFile.WriteLine(line);
            }
            errorsOutputFile.Close();
            headersOutputFile.Close();
        }
    }
}
